#include <config.h>

#include <stdlib.h>
#include <gtk/gtk.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <gdk/gdk.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>

#include <main.h>
#include <error.h>

#include <bmp_sm.h>
#include <bmp_ui.h>
#include <bmp_playlist.h>
#include <bmp_configuration.h>
#include <bmp_play.h>
#include <bmp_resource_manager.h>
#include <bmp_metadata.h>
#include <fam_monitor.h>
#include <list_manager.h>
#include <logger.h>

#ifdef HAVE_GST
#include <gst/gst.h>
#endif

#include <g_object_get.h>
#include <util.h>

gchar *bmp_paths[BMP_N_PATHS];

BmpResourceManager *bmp_rm = NULL;
BmpConfiguration   *bmp_configuration = NULL;
BmpPlaylist        *bmp_playlist = NULL;
BmpPlay		   *bmp_play = NULL;
BmpSystemControl   *bmp_system_control = NULL;
FamMonitor	   *bmp_fam_monitor = NULL;
BmpUI              *bmp_ui = NULL;
BmpMetadataCache   *bmp_metadata_cache = NULL;

static gboolean run_headless  = FALSE;
#ifdef HAVE_DBUS
static gboolean no_remote     = FALSE;
#endif
static gboolean become_daemon = FALSE;
static gboolean log_stdout    = FALSE;

static GOptionEntry bmp_options[] = {
    {"no-log", 0, 0, G_OPTION_ARG_NONE, &log_stdout,
     "Disable logfile, log messages to shell", NULL},
    {"no-ui", 0, 0, G_OPTION_ARG_NONE, &run_headless,
     "Disable UI", NULL},
#ifdef HAVE_DBUS
    {"no-remote", 0, 0, G_OPTION_ARG_NONE, &no_remote,
     "Disable DBUS remote interface", NULL},
#endif
    {"daemon", 0, 0, G_OPTION_ARG_NONE, &become_daemon,
     "Run as a daemon (implies --no-ui)", NULL},
    {NULL}
};

static void
make_directory(const gchar * path, mode_t mode)
{
    if (mkdir(path, mode) == 0)
        return;

    if (errno == EEXIST)
        return;

    g_printerr(Q_("Could not create directory (%s): %s"), path,
               g_strerror(errno));
}

static void
bmp_make_user_dir(void)
{
    const mode_t mode755 = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
  
    make_directory(bmp_paths[BMP_PATH_USER_DIR], mode755);
    make_directory(bmp_paths[BMP_PATH_USER_SKIN_DIR], mode755);
    make_directory(bmp_paths[BMP_PATH_SKIN_THUMB_DIR], mode755);
    make_directory(bmp_paths[BMP_PATH_PLAYLISTS], mode755);
}

static void
bmp_free_paths(void)
{
    int i;

    for (i = 0; i < BMP_N_PATHS; i++)
    {
        g_free(bmp_paths[i]);
        bmp_paths[i] = 0;
    }
}

#define USER_PATH(path) \
    g_build_filename(bmp_paths[BMP_PATH_USER_DIR], path, NULL);

static void
bmp_init_paths(void)
{
    gint i;

    for (i = 0; i < BMP_N_PATHS; i++)
    {
        bmp_paths[i] = NULL;
    }

    bmp_paths[BMP_PATH_USER_DIR] = g_build_filename(g_get_home_dir(), BMP_RCPATH, NULL);
    bmp_paths[BMP_PATH_USER_SKIN_DIR] = USER_PATH(BMP_SKIN_DIR_BASENAME);
    bmp_paths[BMP_PATH_SKIN_THUMB_DIR] = USER_PATH(BMP_SKIN_THUMB_DIR_BASENAME);
    bmp_paths[BMP_PATH_PLAYLISTS] = USER_PATH(BMP_PLAYLISTS_DIR_BASENAME);
    bmp_paths[BMP_PATH_LOG_FILE] = USER_PATH(BMP_LOG_BASENAME);
    bmp_paths[BMP_PATH_METADATA_CACHE] = USER_PATH(BMP_METADATA_CACHE_BASENAME);
#if 0
    bmp_paths[BMP_PATH_CONFIG_FILE] = USER_PATH(BMP_CONFIG_BASENAME);
    bmp_paths[BMP_PATH_ACCEL_FILE] = USER_PATH(BMP_ACCEL_BASENAME);
#endif

    bmp_make_user_dir ();

}

gint
main(gint argc, gchar** argv) {

    gtk_init_with_args (&argc,&argv, "", bmp_options, NULL, NULL);
#ifdef HAVE_GST
    gst_init (&argc, &argv);
#else
    g_thread_init (NULL);
#endif

    gtk_set_locale();
    bindtextdomain(PACKAGE, LOCALEDIR);
    bind_textdomain_codeset(PACKAGE, "UTF-8");
    textdomain(PACKAGE);

    bmp_init_paths ();

    if (! log_stdout)
	bmp_logger_start(bmp_paths[BMP_PATH_LOG_FILE]);

    session_start();

    bmp_rm = 
	bmp_resource_manager_new();
    bmp_configuration  = 
	bmp_configuration_new ();

    /* A small exception so we get the sequence right */
    LOOKUP_OR_CREATE_DEFAULT("bmp", "playlists-path", bmp_paths[BMP_PATH_PLAYLISTS]); 

    if (become_daemon) {
	if (!bmp_become_daemon ())
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Unable to daemonize, proceeding with normal startup.");
	else
		run_headless = TRUE;
    }
  

    bmp_fam_monitor = 
	fam_monitor_new();

#ifdef HAVE_DBUS
    if (bmp_configuration_key_fetch_boolean(bmp_configuration, "bmp", "no-remote")) no_remote = TRUE;
#endif

    bmp_playlist = 
	bmp_playlist_new(); 
    bmp_play = 
	bmp_play_new();

#ifdef HAVE_DBUS
    bmp_system_control = 
	bmp_system_control_new(no_remote);
#else
    bmp_system_control = 
	bmp_system_control_new(FALSE);
#endif

    if (bmp_system_control) {

        bmp_metadata_cache = 
	    bmp_metadata_cache_new ();

	bmp_playlist_load_playlists (bmp_playlist);

        if (bmp_configuration_key_fetch_boolean(bmp_configuration, "bmp", "no-ui")) run_headless = TRUE;

	if (! run_headless) {
	    bmp_system_control_ui_start (bmp_system_control);
        } else {  
	    g_message("%s: Not starting UI on request", G_STRFUNC);
	}

        gtk_main ();
    }

    if (bmp_play)
        bmp_play_stop_stream (bmp_play);

    if (bmp_metadata_cache) {
        g_object_unref (bmp_metadata_cache);
	bmp_metadata_cache = NULL;
    }

    if (bmp_play) {
	g_object_unref (bmp_play);
	bmp_play = NULL;
    }

    if (bmp_playlist) {
	g_object_unref (bmp_playlist);
	bmp_playlist = NULL;
    }

    if (bmp_fam_monitor) {
	g_object_unref (bmp_fam_monitor);
	bmp_fam_monitor = NULL;
    }

    if (bmp_configuration) {
	g_object_unref (bmp_configuration);
	bmp_configuration = NULL;
    }

     if (bmp_system_control) {
	g_object_unref (bmp_system_control);
	bmp_system_control = NULL;
    }
 
    bmp_logger_stop(); 

    session_end();

    exit (EXIT_SUCCESS);
}
